import pathlib
from collections.abc import Iterator
from typing import TYPE_CHECKING

from babel.support import NullTranslations, Translations

from fastlife.shared_utils.resolver import resolve_path

if TYPE_CHECKING:
    from fastlife import Request  # coverage: ignore

locale_name = str


def find_mo_files(root_path: str) -> Iterator[tuple[str, str, pathlib.Path]]:
    root = pathlib.Path(root_path)

    # Walk through the directory structure and match the pattern
    for locale_dir in root.iterdir():
        if locale_dir.is_dir():  # Ensure it's a directory (locale)
            lc_messages_dir = locale_dir / "LC_MESSAGES"
            if lc_messages_dir.exists() and lc_messages_dir.is_dir():
                for mo_file in lc_messages_dir.glob("*.mo"):  # Find .mo files
                    yield locale_dir.name, mo_file.name[:-3], mo_file


class Localizer:
    def __init__(
        self, request: "Request", translations: Translations | NullTranslations
    ) -> None:
        self.locale_name = request.locale_name
        self.translations = translations

    def gettext(self, message: str, mapping: dict[str, str] | None = None) -> str:
        ret = self.translations.gettext(message)
        if mapping is not None:
            ret = ret.format(**mapping)
        return ret

    def ngettext(
        self, singular: str, plural: str, n: int, mapping: dict[str, str] | None = None
    ) -> str:
        ret = self.translations.ngettext(singular, plural, n)
        mapping_num = {"num": n, **(mapping or {})}
        return ret.format(**mapping_num)

    def dgettext(
        self, domain: str, message: str, mapping: dict[str, str] | None = None
    ) -> str:
        ret = self.translations.dgettext(domain, message)
        if mapping:
            ret = ret.format(**mapping)
        return ret

    def dngettext(
        self,
        domain: str,
        singular: str,
        plural: str,
        n: int,
        mapping: dict[str, str] | None = None,
    ) -> str:
        ret = self.translations.dngettext(domain, singular, plural, n)
        mapping_num = {"num": n, **(mapping or {})}
        return ret.format(**mapping_num)

    def pgettext(
        self, context: str, message: str, mapping: dict[str, str] | None = None
    ) -> str:
        ret = str(self.translations.pgettext(context, message))
        if mapping:
            ret = ret.format(**mapping)
        return ret

    def dpgettext(
        self,
        domain: str,
        context: str,
        message: str,
        mapping: dict[str, str] | None = None,
    ) -> str:
        ret = str(self.translations.dpgettext(domain, context, message))
        if mapping:
            ret = ret.format(**mapping)
        return ret

    def dnpgettext(
        self,
        domain: str,
        context: str,
        singular: str,
        plural: str,
        n: int,
        mapping: dict[str, str] | None = None,
    ) -> str:
        ret = self.translations.dnpgettext(domain, context, singular, plural, n)
        mapping_num = {"num": n, **(mapping or {})}
        return ret.format(**mapping_num)


class LocalizerFactory:
    """Initialize the proper translation context per request."""

    _translations: dict[locale_name, Translations]

    def __init__(self) -> None:
        self._translations = {}

    def load(self, path: str) -> None:
        """
        load a translations from the model.
        :param path: a python module and the locales dir separated by a `:`
        """
        root_path = resolve_path(path)
        for locale_name, domain, file_ in find_mo_files(root_path):
            with file_.open("rb") as f:
                t = Translations(f, domain)
                if locale_name not in self._translations:
                    self._translations[locale_name] = Translations()
                self._translations[locale_name].add(t)
                self._translations[locale_name].merge(t)

    def __call__(self, request: "Request") -> Localizer:
        """Create the translation context for the given request."""
        trans: Translations | NullTranslations | None = self._translations.get(
            request.locale_name
        )
        if not trans:
            trans = self._translations.get(request.registry.settings.default_locale)
        if not trans:
            trans = NullTranslations()
        return Localizer(request, trans)



"""
Widget for field of type Set.
"""

from collections.abc import Sequence

from pydantic import BaseModel, Field

from .base import Widget


class Checkable(BaseModel):
    """A checkable field from a checklist."""

    label: str
    name: str
    value: str
    token: str
    checked: bool
    error: str | None = Field(default=None)

    @property
    def id(self) -> str:
        id = f"{self.name}-{self.value}-{self.token}"
        return id.replace(".", "-").replace("_", "-")

    @property
    def field_name(self) -> str:
        return f"{self.name}[]"


class ChecklistWidget(Widget[Sequence[Checkable]]):
    """
    Widget for field of type Set.

    :param name: field name.
    :param title: title for the widget.
    :param hint: hint for human.
    :param aria_label: html input aria-label value.
    :param value: current value.
    :param error: error of the value if any.
    :param removable: display a button to remove the widget for optional fields.
    :param token: token used to get unique id on the form.
    """

    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        hint: str | None = None,
        aria_label: str | None = None,
        value: Sequence[Checkable],
        error: str | None = None,
        token: str,
        removable: bool,
    ) -> None:
        super().__init__(
            name,
            value=value,
            error=error,
            token=token,
            title=title,
            hint=hint,
            aria_label=aria_label,
            removable=removable,
        )

    def get_template(self) -> str:
        return "pydantic_form.Checklist.jinja"



"""
Base class to of the template renderer.

Fastlife comes with {class}`fastlife.templating.renderer.jinjax.JinjaxTemplateRenderer`,
the rendering engine, it can be overriden from the setting
:attr:`fastlife.config.settings.Settings.template_renderer_class`.

In that case, those base classes have to be implemented.

"""

import abc
from collections.abc import Callable, Mapping
from typing import Any

from fastlife import Request, Response
from fastlife.security.csrf import create_csrf_token

TemplateParams = Mapping[str, Any]


class AbstractTemplateRenderer(abc.ABC):
    """
    An object that will be initialized by an AbstractTemplateRendererFactory,
    passing the request to process.
    """

    request: Request
    """Associated request that needs a response."""

    def __init__(self, request: Request) -> None:
        self.request = request

    @property
    def route_prefix(self) -> str:
        """Used to buid pydantic form widget that do ajax requests."""
        return self.request.registry.settings.fastlife_route_prefix

    def render(
        self,
        template: str,
        *,
        status_code: int = 200,
        content_type: str = "text/html",
        globals: Mapping[str, Any] | None = None,
        params: TemplateParams,
        _create_csrf_token: Callable[..., str] = create_csrf_token,
    ) -> Response:
        """
        Render the template and build the HTTP Response.
        """
        request = self.request
        reg = request.registry
        request.scope[reg.settings.csrf_token_name] = (
            request.cookies.get(reg.settings.csrf_token_name) or _create_csrf_token()
        )
        data = self.render_template(template, **params)
        resp = Response(
            data, status_code=status_code, headers={"Content-Type": content_type}
        )
        resp.set_cookie(
            reg.settings.csrf_token_name,
            request.scope[reg.settings.csrf_token_name],
            secure=request.url.scheme == "https",
            samesite="strict",
            max_age=60 * 15,
        )
        return resp

    @abc.abstractmethod
    def render_template(
        self,
        template: str,
        *,
        globals: Mapping[str, Any] | None = None,
        **params: Any,
    ) -> str:
        """
        Render the given template with the given params.

        While rendering templates, the globals parameter is keps by the instantiated
        renderer and sent to every rendering made by the request.
        This is used by the pydantic form method that will render other templates
        for the request.
        In traditional frameworks, only one template is rendered containing the whole
        pages. But, while rendering a pydantic form, every field is rendered in its
        distinct template. The template renderer keep the globals and git it back
        to every templates. This can be used to fillout options in a select without
        performing an ajax request for example.

        :param template: name of the template to render.
        :param globals: some variable that will be passed to all rendered templates.
        :param params: paramaters that are limited to the main rendered templates.
        :return: The template rendering result.
        """


class AbstractTemplateRendererFactory(abc.ABC):
    """
    The template render factory.

    The implementation of this class is found using the settings
    :attr:`fastlife.config.settings.Settings.template_renderer_class`.
    """

    @abc.abstractmethod
    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """
        While processing an HTTP Request, a renderer object is created giving
        isolated context per request.

        :param Request: the HTTP Request to process.
        :return: The renderer object that will process that request.
        """



"""
Template rending based on JinjaX.
"""

import ast
import logging
import re
import textwrap
from collections.abc import Iterator, Mapping, MutableMapping, Sequence
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    cast,
)

from jinja2 import Template
from jinja2.exceptions import TemplateSyntaxError
from jinjax import InvalidArgument
from jinjax.catalog import Catalog
from jinjax.component import RX_ARGS_START, RX_META_HEADER, Component
from jinjax.exceptions import DuplicateDefDeclaration
from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife import Request
from fastlife.adapters.jinjax.widgets.factory import WidgetFactory
from fastlife.request.form import FormModel
from fastlife.request.localizer import get_localizer

if TYPE_CHECKING:
    from fastlife.config.settings import Settings  # coverage: ignore

from fastlife.services.templates import (
    AbstractTemplateRenderer,
    AbstractTemplateRendererFactory,
)
from fastlife.shared_utils.resolver import resolve, resolve_path

log = logging.getLogger(__name__)

RX_DOC_START = re.compile(r"{#-?\s*doc\s+")
RX_CONTENT = re.compile(r"\{\{-?\s*content\s*-?\}\}", re.DOTALL)
RX_COMMENT_REPLACE = re.compile(r"{#[^#]+#}")


def has_content(source: str) -> bool:
    nocomment = RX_COMMENT_REPLACE.sub("", source)
    return len(RX_CONTENT.findall(nocomment)) > 0


def generate_docstring(
    func_def: ast.FunctionDef, component_name: str, add_content: bool
) -> str:
    """Generate a docstring for the component."""
    # Extract function name and docstring
    docstring = (ast.get_docstring(func_def, clean=True) or "").strip()
    if docstring:
        docstring = textwrap.dedent(docstring)
        docstring_lines = [line for line in docstring.split("\n")]
        # Add a newline for separation after the function docstring
        docstring_lines.append("")
    else:
        docstring_lines = []

    component_params: list[str] = []

    # Function for processing an argument and adding its docstring lines
    def process_arg(arg: ast.arg, default_value: Any = None) -> None:
        arg_name = arg.arg
        param_desc = ""
        # Extract the type annotation (if any)
        if (
            isinstance(arg.annotation, ast.Subscript)
            and isinstance(arg.annotation.value, ast.Name)
            and arg.annotation.value.id == "Annotated"
        ):
            # For Annotated types, we expect the first argument to be the type and
            # the second to be the description
            type_annotation = arg.annotation.slice.elts[0]  # type: ignore
            param_type = ast.unparse(type_annotation)  # type: ignore

            if len(arg.annotation.slice.elts) > 1 and isinstance(  # type: ignore
                arg.annotation.slice.elts[1],  # type: ignore
                ast.Constant,  # type: ignore
            ):
                param_desc = arg.annotation.slice.elts[1].value  # type: ignore
        else:
            # Otherwise, just use the type if available
            param_type = ast.unparse(arg.annotation) if arg.annotation else "Any"

        # Build the parameter docstring line
        docstring_lines.append(f":param {arg_name.rstrip('_')}: {param_desc}".strip())

        # Build the string representation of the parameter
        param_str = f"{arg_name}: {param_type}"
        if default_value is not None:
            param_str += f" = {ast.unparse(default_value)}"  # type: ignore

        component_params.append(param_str)

    # Process keyword-only arguments
    kwonlyargs = func_def.args.kwonlyargs
    kw_defaults = func_def.args.kw_defaults

    for arg, default in zip(kwonlyargs, kw_defaults, strict=False):
        process_arg(arg, default)

    if add_content:
        component_params.append("content: Any")
        docstring_lines.append(":param content: child node.")

    return (
        f"{component_name}({', '.join(component_params)})"
        + "\n"
        + "\n    "
        + ("\n    ".join(docstring_lines).strip()).replace("\n    \n", "\n\n")
        + "\n"
    )


class InspectableComponent(Component):
    """
    JinjaX Components override that support intropspection.

    The Component class of JinjaX does not support documentation by default.
    """

    __slots__ = (
        "name",
        "prefix",
        "url_prefix",
        "required",
        "optional",
        "css",
        "js",
        "path",
        "mtime",
        "tmpl",
        "source",
    )

    def __init__(
        self,
        *,
        name: str,
        prefix: str = "",
        url_prefix: str = "",
        source: str = "",
        mtime: float = 0,
        tmpl: "Template | None" = None,
        path: "Path | None" = None,
    ) -> None:
        super().__init__(
            name=name,
            prefix=prefix,
            url_prefix=url_prefix,
            source=source,
            mtime=mtime,
            tmpl=tmpl,
            path=path,
        )
        self.source = source

    def as_def(self) -> ast.FunctionDef:
        """Return the component definition as a python function definition."""
        signature = "def component(): pass"
        match = RX_META_HEADER.match(self.source)
        if match:
            headers = match.group(0)
            header = headers.split("#}")[:-1]
            def_found = False
            docstring = ""

            expr = None
            while header:
                item = header.pop(0).strip(" -\n")

                expr = self.read_metadata_item(item, RX_ARGS_START)  # type: ignore
                if expr:
                    if def_found:
                        raise DuplicateDefDeclaration(self.name)
                    def_found = True
                    continue

                doc = self.read_metadata_item(item, RX_DOC_START)  # type: ignore
                if doc:
                    docstring += f"    {doc.strip()}\n"
                    continue

            if expr:
                signature = f"""\
def component(*, {expr}):
    '''
    {docstring or ""}
    '''
    ...
"""
            elif docstring:
                signature = f"""\
def component():
    '''
    {docstring}
    '''
    ...
"""

        try:
            astree = ast.parse(signature)
        except SyntaxError as err:
            raise InvalidArgument(err) from err
        return cast(ast.FunctionDef, astree.body[0])

    def build_docstring(self) -> str:
        """Build a rst docstring for the jinjax component."""
        func_def = self.as_def()
        prefix = f"{self.prefix}." if self.prefix else ""
        ret = ".. jinjax:component:: " + generate_docstring(
            func_def, f"{prefix}{self.name}", has_content(self.source)
        )
        return ret


class InspectableCatalog(Catalog):
    """
    JinjaX Catalog with introspection support.

    Override the catalog in order to iterate over components to build the doc.
    """

    def iter_components(
        self,
        ignores: Sequence[re.Pattern[str]] | None = None,
        includes: Sequence[re.Pattern[str]] | None = None,
    ) -> Iterator[InspectableComponent]:
        """
        Walk into every templates from the settings, iterate over components.

        :params ignores: filter components using an block list.
        :params includes: filter components using an allowed list.
        """
        for prefix, loader in self.prefixes.items():
            for t in loader.list_templates():
                name, file_ext = t.split(".", maxsplit=1)
                name = name.replace("/", ".")
                path, tmpl_name = self._get_component_path(
                    prefix, name, file_ext=file_ext
                )

                to_include = True
                if includes:
                    to_include = False
                    for include in includes:
                        if include.match(name):
                            to_include = True
                            break
                if to_include and ignores:
                    for ignore in ignores:
                        if ignore.match(name):
                            to_include = False
                            break

                if to_include:
                    component = InspectableComponent(
                        name=name, prefix=prefix, path=path, source=path.read_text()
                    )

                    self.jinja_env.loader = loader
                    try:
                        component.tmpl = self.jinja_env.get_template(
                            tmpl_name, globals=self._tmpl_globals
                        )
                    except TemplateSyntaxError as exc:
                        log.error(f"Syntax Error: {exc} on {exc.lineno} :")
                        log.error(path.read_text())
                        continue
                    yield component


def build_searchpath(template_search_path: str) -> Sequence[str]:
    """
    Build the path containing templates.

    Path may be absolute directories or directories relative to a python
    package. For instance, the `fastlife:components` is the directory components
    inside the fastlife installation dir.

    :param template_search_path: list of path separated by a comma (`,`).
    :return: List resolved path.
    """
    searchpath: list[str] = []
    paths = template_search_path.split(",")

    for path in paths:
        if ":" in path:
            searchpath.append(resolve_path(path))
        else:
            searchpath.append(path)
    return searchpath


class JinjaxRenderer(AbstractTemplateRenderer):
    """Render templates using JinjaX."""

    def __init__(
        self,
        catalog: InspectableCatalog,
        request: Request,
    ):
        super().__init__(request)
        self.catalog = catalog
        self.settings = request.registry.settings
        self.globals: MutableMapping[str, Any] = {}
        self.translations = get_localizer(request)

    def build_globals(self) -> Mapping[str, Any]:
        """
        Build globals variables accessible in any templates.

        * `request` is the {class}`current request <fastlife.request.request.Request>`
        * `csrf_token` is used to build for {jinjax:component}`CsrfToken`.
        """
        settings = self.settings
        ret = {
            "request": self.request,
            "csrf_token": {
                "name": settings.csrf_token_name,
                "value": self.request.scope.get(settings.csrf_token_name, ""),
            },
            "pydantic_form": self.pydantic_form,
            "localizer": self.translations,
            **self.globals,
        }
        return ret

    def render_template(
        self,
        template: str,
        *,
        globals: Mapping[str, Any] | None = None,
        **params: Any,
    ) -> str:
        # Jinja template does accept the file extention while rendering the template
        # we strip it before rendering.
        template = template[: -len(self.settings.jinjax_file_ext) - 1]
        if globals:
            self.globals.update(globals)

        self.catalog.jinja_env.install_gettext_translations(  # type: ignore
            self.translations, newstyle=True
        )

        return self.catalog.render(  # type: ignore
            template, __globals=self.build_globals(), **params
        )

    def pydantic_form(
        self, model: FormModel[Any], *, token: str | None = None
    ) -> Markup:
        return WidgetFactory(self, token).get_markup(model)

    def pydantic_form_field(
        self,
        model: type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        return (
            WidgetFactory(self, token)
            .get_widget(
                model,
                form_data={},
                form_errors={},
                prefix=(name or self.request.registry.settings.form_data_model_prefix),
                removable=removable,
                field=field,
            )
            .to_html(self)
        )


class JinjaxTemplateRenderer(AbstractTemplateRendererFactory):
    """
    The default template renderer factory. Based on JinjaX.

    :param settings: setting used to configure jinjax.
    """

    def __init__(self, settings: "Settings") -> None:
        globals = resolve(settings.jinjax_global_catalog_class)().model_dump()

        self.catalog = InspectableCatalog(
            use_cache=settings.jinjax_use_cache,
            file_ext=settings.jinjax_file_ext,
            auto_reload=settings.jinjax_auto_reload,
            globals=globals,
        )
        self.catalog.jinja_env.add_extension("jinja2.ext.i18n")
        for path in build_searchpath(settings.template_search_path):
            self.catalog.add_folder(path)

    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """Build the renderer to render request for template."""
        return JinjaxRenderer(
            self.catalog,
            request,
        )



"""
Widget for field of type Enum or Literal.
"""

from collections.abc import Sequence

from .base import Widget


class DropDownWidget(Widget[str]):
    """
    Widget for field of type Enum or Literal.

    :param name: field name.
    :param title: title for the widget.
    :param hint: hint for human.
    :param aria_label: html input aria-label value.
    :param value: current value.
    :param error: error of the value if any.
    :param options: List of possible values.
    :param removable: display a button to remove the widget for optional fields.
    :param token: token used to get unique id on the form.
    """

    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        hint: str | None = None,
        aria_label: str | None = None,
        value: str | None = None,
        error: str | None = None,
        options: Sequence[tuple[str, str]] | Sequence[str],
        removable: bool = False,
        token: str | None = None,
    ) -> None:
        super().__init__(
            name,
            value=value,
            error=error,
            title=title,
            token=token,
            removable=removable,
            hint=hint,
            aria_label=aria_label,
        )
        self.options: list[dict[str, str]] = []
        for opt in options:
            if isinstance(opt, tuple):
                self.options.append({"value": opt[0], "text": opt[1]})
            else:
                self.options.append({"value": opt, "text": opt})

    def get_template(self) -> str:
        return "pydantic_form.Dropdown.jinja"



"""
Bind template to the view in order to build an html response.
"""

from collections.abc import Callable
from typing import Any

from fastapi import Depends, Response

from fastlife.request import Request
from fastlife.security.csrf import create_csrf_token

Template = Callable[..., Response]
"""Type to annotate a FastAPI depency injection."""

TemplateEngine = Callable[[Request], Template]


def get_template(template: str, *, content_type: str = "text/html") -> TemplateEngine:
    """
    Return a closure to render the given template.

    :param template: path to template to render.
    :param content_type: response ``Content-Type`` header.
    """

    def render_template(
        request: Request,
        *,
        _create_csrf_token: Callable[..., str] = create_csrf_token,
    ) -> Template:
        reg = request.registry

        def parametrizer(**kwargs: Any) -> Response:
            return reg.get_renderer(template)(request).render(
                template, content_type=content_type, params=kwargs
            )

        return parametrizer

    return render_template


def template(template_path: str) -> Template:
    """
    Return a FastAPI dependency template engine ready to render the template.


    :param template_path: path to template to render by the engine setup in the regitry.
    :return: A callable accepting kwargs to pass as the context, returning a string.
    """
    return Depends(get_template(template_path))



from typing import Any

from .base import Widget


class HiddenWidget(Widget[str]):
    '''
    Widget to annotate to display a field as an hidden field.

    ::
        from pydantic import BaseModel
        from fastlife.adapters.jinjax.widgets.hidden import HiddenWidget

        class MyForm(BaseModel):
            id: Annotated[str, HiddenWidget] = Field(...)
            """Identifier in the database."""

    '''

    def __init__(
        self,
        name: str,
        *,
        value: str,
        token: str,
        **kwargs: Any,
    ) -> None:
        super().__init__(name, value=value, token=token)

    def get_template(self) -> str:
        return "pydantic_form.Hidden.jinja"



from collections.abc import Sequence
from typing import Any

from markupsafe import Markup

from fastlife.services.templates import AbstractTemplateRenderer

from .base import TypeWrapper, Widget


class SequenceWidget(Widget[Sequence[Widget[Any]]]):
    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        hint: str | None = None,
        aria_label: str | None = None,
        value: Sequence[Widget[Any]] | None,
        error: str | None = None,
        item_type: type[Any],
        token: str,
        removable: bool,
    ):
        super().__init__(
            name,
            value=value,
            error=error,
            title=title,
            hint=hint,
            aria_label=aria_label,
            token=token,
            removable=removable,
        )
        self.item_type = item_type

    def get_template(self) -> str:
        return "pydantic_form.Sequence.jinja"

    def build_item_type(self, route_prefix: str) -> TypeWrapper:
        return TypeWrapper(self.item_type, route_prefix, self.name, self.token)

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version."""
        children = [Markup(item.to_html(renderer)) for item in self.value or []]
        return Markup(
            renderer.render_template(
                self.get_template(),
                widget=self,
                type=self.build_item_type(renderer.route_prefix),
                children_widgets=children,
            )
        )



from collections.abc import Sequence

from .base import Widget


class TextWidget(Widget[str]):
    """
    Widget for text like field (email, ...).

    :param name: input name.
    :param title: title for the widget.
    :param hint: hint for human.
    :param aria_label: html input aria-label value.
    :param placeholder: html input placeholder value.
    :param error: error of the value if any.
    :param value: current value.
    :param removable: display a button to remove the widget for optional fields.
    :param token: token used to get unique id on the form.
    """

    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        hint: str | None = None,
        aria_label: str | None = None,
        placeholder: str | None = None,
        error: str | None = None,
        value: str = "",
        input_type: str = "text",
        removable: bool = False,
        token: str,
    ) -> None:
        super().__init__(
            name,
            value=value,
            title=title,
            hint=hint,
            aria_label=aria_label,
            token=token,
            error=error,
            removable=removable,
        )
        self.placeholder = placeholder or ""
        self.input_type = input_type

    def get_template(self) -> str:
        return "pydantic_form.Text.jinja"


class TextareaWidget(Widget[Sequence[str]]):
    """
    Render a Textearea for a string or event a sequence of string.

    ```
    from fastlife.adapters.jinjax.widgets.text import TextareaWidget
    from pydantic import BaseModel, Field, field_validator

    class TaggedParagraphForm(BaseModel):
        paragraph: Annotated[str, TextareaWidget] = Field(...)
        tags: Annotated[Sequence[str], TextareaWidget] = Field(
            default_factory=list,
            title="Tags",
            description="One tag per line",
        )

        @field_validator("tags", mode="before")
        def split(cls, s: Any) -> Sequence[str]:
            return s.split() if s else []
    ```

    :param name: input name.
    :param title: title for the widget.
    :param hint: hint for human.
    :param aria_label: html input aria-label value.
    :param placeholder: html input placeholder value.
    :param error: error of the value if any.
    :param value: current value.
    :param removable: display a button to remove the widget for optional fields.
    :param token: token used to get unique id on the form.

    """

    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        hint: str | None = None,
        aria_label: str | None = None,
        placeholder: str | None = None,
        error: str | None = None,
        value: Sequence[str] | None = None,
        removable: bool = False,
        token: str,
    ) -> None:
        super().__init__(
            name,
            value=value or [],
            title=title,
            hint=hint,
            aria_label=aria_label,
            token=token,
            error=error,
            removable=removable,
        )
        self.placeholder = placeholder or ""

    def get_template(self) -> str:
        return "pydantic_form.Textarea.jinja"



"""
Widget for field of type Union.
"""

from collections.abc import Sequence
from typing import Any, Union

from markupsafe import Markup
from pydantic import BaseModel

from fastlife.services.templates import AbstractTemplateRenderer

from .base import TypeWrapper, Widget


class UnionWidget(Widget[Widget[Any]]):
    """
    Widget for union types.

    :param name: input name.
    :param title: title for the widget.
    :param hint: hint for human.
    :param aria_label: html input aria-label value.
    :param value: current value.
    :param error: error of the value if any.
    :param children_types: childrens types list.
    :param removable: display a button to remove the widget for optional fields.
    :param token: token used to get unique id on the form.

    """

    def __init__(
        self,
        name: str,
        *,
        title: str | None,
        hint: str | None = None,
        aria_label: str | None = None,
        value: Widget[Any] | None,
        error: str | None = None,
        children_types: Sequence[type[BaseModel]],
        removable: bool = False,
        token: str,
    ):
        super().__init__(
            name,
            value=value,
            error=error,
            title=title,
            hint=hint,
            aria_label=aria_label,
            token=token,
            removable=removable,
        )
        self.children_types = children_types
        self.parent_name = name

    def build_types(self, route_prefix: str) -> Sequence[TypeWrapper]:
        """Wrap types in the union in order to get the in their own widgets."""
        return [
            TypeWrapper(typ, route_prefix, self.name, self.token)
            for typ in self.children_types
        ]

    def get_template(self) -> str:
        return "pydantic_form.Union.jinja"

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version."""
        child = Markup(self.value.to_html(renderer)) if self.value else ""
        return Markup(
            renderer.render_template(
                self.get_template(),
                widget=self,
                types=self.build_types(renderer.route_prefix),
                parent_type=TypeWrapper(
                    Union[tuple(self.children_types)],  # type: ignore # noqa: UP007
                    renderer.route_prefix,
                    self.parent_name,
                    self.token,
                    title=self.title,
                ),
                child=child,
            )
        )
